"""
The module represents the geometric primitives for layout design.

name: for retrieval [unique]
label: for display
layer:
- center
- axis
- extremes
- region
- bounding box

styles are managed by canvas

Curves are parametric curves. They are defined by a start point, a tangent and a list of segments. The segments are:
   - lineto: end
   - arcto: radius, end
   - lineby: length
   - bendby: length, angle, roll

Curve query:
    - point: return the point at a given accumulated path length s
    - points: return a set of points covering the curve at a give accuracy
    - tangent: return the tangent at a given accumulated path length s

Primitive api:

points(): return points
mesh(): return mesh
 
"""

import numpy as np
from scipy.spatial.transform import Rotation, Slerp

from .pose import Element, Pose


class Point:
    def __init__(self, point=None, name=None, label=None, layer=None):
        """
        Point shape must be (3,) or (4,)
        """
        self.arr = np.ones(4, dtype=float)
        if point is None:
            self.arr[:3] = 0
        else:
            self.arr[:3] = point[:3]
        self.name = name
        self.label = label
        self.layer = layer

    def loc(self):
        return self.arr[:3]

    def __repr__(self):
        args = []
        if self.name is not None:
            args.append(f"name={self.name}")
        return f"Point({self.arr[:3]},{', '.join(args)})"


class Points:
    def __init__(self, points, name=None, label=None, layer=None):
        """
        stored in 4xN vector with 1 in the last row
        """
        if hasattr(points, "shape"):
            if points.shape[0] == 3:
                self.arr = np.ones((4, points.shape[1]), dtype=float)
                self.arr[:3] = points
            elif points.shape[0] == 4:
                self.arr = points
            else:
                raise ValueError("Points shape must be (3,N) or (4,N)")
        self.name = name
        self.label = label
        self.layer = layer

    def __repr__(self):
        args = []
        if self.name is not None:
            args.append(f"name={self.name}")
        return f"<Points({self.arr[:3]},{', '.join(args)})>"


class Line(Element):
    def __init__(self, start, end, name=None, label=None, layer=None):
        if not isinstance(start, Pose):
            start = Pose(*start)
        if not isinstance(end, Pose):
            end = Pose(*end)
        self.start = start
        self.end = end
        self.name = name
        self.label = label
        self.layer = layer

    def __repr__(self):
        args = []
        if self.name is not None:
            args.append(f"name={self.name}")
        return f"Line({self.start},{self.end},{', '.join(args)})"

    def length(self):
        return self.start.distance(self.end)

    def point(self, s):
        pstart = self.start.loc
        pend = self.end.loc
        rstart = Rotation.from_matrix(self.start.rot)
        rend = Rotation.from_matrix(self.end.rot)
        rot = Slerp([0, self.length], [rstart, rend])(s)
        return Pose(loc=pstart + s * (pend - pstart), rot=rot.as_matrix())

    def points(self):
        return np.array([self.start.loc4, self.end.loc4]).T


class Rectangle(Element):
    def __init__(self, name=None, lx=1, ly=1, label=None, layer=None):
        self.lx = lx
        self.ly = ly
        self.name = name
        self.label = label
        self.layer = layer

    def __repr__(self):
        args = []
        if self.name is not None:
            args.append(f"name={self.name}")
        return f"Rectangle(lx={self.lx},lx={self.ly})"

    @property
    def center(self):
        return Pose(name=f"{self.name}/center")

    @property
    def ul(self):
        return Pose(name=f"{self.name}/ul", x=-self.lx / 2, y=self.ly / 2)

    @property
    def ur(self):
        return Pose(name=f"{self.name}/ur", x=self.lx / 2, y=self.ly / 2)

    @property
    def ll(self):
        return Pose(name=f"{self.name}/ll", x=-self.lx / 2, y=-self.ly / 2)

    @property
    def lr(self):
        return Pose(name=f"{self.name}/lr", x=self.lx / 2, y=-self.ly / 2)

    @property
    def left(self):
        return Pose(name=f"{self.name}/left", x=-self.lx / 2)

    @property
    def right(self):
        return Pose(name=f"{self.name}/right", x=self.lx / 2)

    @property
    def top(self):
        return Pose(name=f"{self.name}/top", y=self.ly / 2)

    @property
    def bottom(self):
        return Pose(name=f"{self.name}/bottom", y=-self.ly / 2)

    def points(self):
        return np.array(
            [
                pp.loc4
                for pp in (self.ul, self.ur, self.lr, self.ll, self.ul)
            ]
        )

    def render(self, style):
        primitives = []
        points = np.array(
            [
                pp.loc4
                for pp in (self.ul, self.ur, self.lr, self.ll, self.ul)
            ]
        ).T
        print(points)
        primitives.append(Polygon(points=points).at(name=self.name))
        return primitives


class Polyline(Element):
    def __init__(self, name=None, points=None, label=None, layer=None):
        self.name = name
        self.points = points
        self.label = label
        self.layer = layer

    def __repr__(self):
        return f"Polyline({self.points})"



class Polygon(Polyline):
    def __init__(self, name=None, points=None, label=None, layer=None):
        self.name = name
        self.points = points
        self.label = label
        self.layer = layer

    def __repr__(self):
        return f"Polygon({self.points})"


class Bend:
    """
    Defined by starting pose:
       plane: in xz rolled by roll angle
       angle: angle spanned by the arc
       length: length of the arc

    """

    def __init__(
        self, start, length, angle, roll=0, name=None, label=None, layer=None
    ):
        self.start = start
        self.angle = angle
        self.length = length
        self.roll = roll
        self.name = name
        self.label = label
        self.layer = layer

    def __repr__(self):
        args = []
        if self.name is not None:
            args.append(f"name={self.name}")
        if self.roll != 0:
            args.append(f"roll={self.roll}")
        return (
            f"Bend({self.start},{self.length},{self.angle},{', '.join(args)})"
        )

    def point(self, s):
        # using mad-x formula
        fullangle = np.deg2rad(self.angle)
        radius = self.length / fullangle
        alpha = fullangle * s / self.length
        ca = np.cos(alpha)
        sa = np.sin(alpha)
        psi = np.deg2rad(self.roll)
        cp = np.cos(psi)
        sp = np.sin(psi)
        R = np.array([radius * (ca - 1), 0, radius * sa])
        S = np.array([[ca, 0, -sa], [0, 1, 0], [sa, 0, ca]])
        T = np.array([[cp, -sp, 0], [sp, cp, 0], [0, 0, 1]])
        Ti = np.array([[cp, sp, 0], [-sp, cp, 0], [0, 0, 1]])
        R = T @ R
        S = T @ S @ Ti
        loc = self.start.rot @ R + self.start.loc
        rot = self.start.rot @ S
        return Point(loc=loc, rot=rot)

    def points(self, steps=5):
        s = np.linspace(0, self.length, steps)
        return [self.point(si) for si in s]

    @property
    def end(self):
        return self.point(self.length)


class Ellipse:
    def __init__(self, center, radius_x, radius_y):
        self.center = center
        self.radius_x = radius_x
        self.radius_y = radius_y


class Circle:
    def __init__(self, center, radius):
        self.center = center
        self.radius = radius


# curve segments
class LineTo:
    def __init__(self, end):
        self.end = end

    def __repr__(self):
        return f"LineTo({self.end})"

    def get_segment(self, start):
        length = start.distance(self.end)
        return Line(start, self.end), length, self.end


class ArcTo:
    def __init__(self, radius, end):
        self.radius = radius
        self.end = end

    def __repr__(self):
        return f"ArcTo({self.radius},{self.end})"

    def get_segment(self, start):
        raise NotImplementedError


class LineBy:
    def __init__(self, length, axis="z"):
        self.length = length
        self.axis = axis

    def __repr__(self):
        args = []
        args.append(f"length={self.length}")
        if self.axis != "z":
            args.append(f"axis={self.axis}")
        return f"LineBy({','.join(args)})"

    def get_segment(self, start):
        end = getattr(start.new(), f"t{self.axis}")(self.length)
        return Line(start, end), self.length, end


class BendBy:
    def __init__(self, length=0, angle=0, roll=0, axis="xz"):
        self.length = length
        self.angle = angle
        self.roll = roll
        self.axis = axis

    def __repr__(self) -> str:
        args = []
        if self.length != 0:
            args.append(f"length={self.length}")
        if self.angle != 0:
            args.append(f"angle={self.angle}")
        if self.roll != 0:
            args.append(f"roll={self.roll}")
        if self.axis != "z":
            args.append(f"axis={self.axis}")
        return f"BendBy({','.join(args)})"

    def get_segment(self, start):
        segment = Bend(start, self.length, self.angle, self.roll)
        return segment, self.length, segment.end


class Curve:
    """
    A smooth curve is a parametric curve defined by a start pose and a list of segments.

    The curve is smooth is all points and tangents are continuous.

    Each s is associated to one and one-only pose.

    """

    @classmethod
    def from_svgpath(cls, svgpath):
        pass

    def __init__(self, start=None, specs=None, s_start=0.0, lookup_ds=1.0):
        if start is None:
            start = Point()
        if specs is None:
            specs = []
        self.start = start
        self.end = start
        self.s_start = s_start
        self.s_end = s_start
        self.specs = specs
        self.segments = []
        self.lookup_ds = lookup_ds
        self.lookup_end = s_start
        self.lookup = []
        for spec in specs:
            self.add_spec(spec)

    def add_spec(self, spec):
        seg_start = self.s_end
        segment, length, end = spec.get_segment(self.end)
        self.s_end += length
        self.end = end
        self.segments.append((seg_start, segment))
        idx = len(self.segments) - 1
        self.specs.append(spec)
        steps = int((self.s_end - self.lookup_end) / self.lookup_ds)
        self.lookup += [idx] * steps
        self.lookup_end += steps * self.lookup_ds
        self.length = self.s_end - self.s_start

    def point(self, s):
        if s < self.s_start or s > self.s_end:
            raise ValueError(
                f"Curve point out of range {s} not in [{self.s_start},{self.s_end}]"
            )
        lookup_idx = int((s - self.s_start) // self.lookup_ds)
        seg_idx = self.lookup[lookup_idx]
        segment_s, segment = self.segments[seg_idx]
        if s < segment_s:
            segment_s, segment = self.segments[seg_idx - 1]
        return segment.point(s - segment_s)

    def lineto(self, end):
        self.add_spec(LineTo(end))

    def arcto(self, radius, end):
        self.add_spec(ArcTo(radius, end))

    def lineby(self, length, axis="z"):
        self.add_spec(LineBy(length, axis))

    def bendby(self, length=0, angle=0, roll=0, axis="z"):
        self.add_spec(BendBy(length, angle, roll, axis))

    def points(self, steps=5):
        points = []
        for _, segment in self.segments:
            points += segment.points()
        return points

    def tangent(self, s):
        pass

    def to_svgpath(self):
        pass

    def __repr__(self):
        return f"Curve: {self.s_start}, {self.s_end}, {len(self.segments)} segments"


class Box:
    def __init__(self, center, size, name=None, label=None, layer=None):
        self.center = center
        self.size = size
        self.name = name
        self.label = label
        self.layer = layer


class Tube:
    def __init__(self, curve, sections, name=None, label=None, layer=None):
        self.curve = curve
        self.sections = sections
        self.name = name
        self.label = label
        self.layer = layer


class Text:
    def __init__(self, text, name=None, label=None, layer=None):
        self.text = text
        self.name = name
        self.label = label
        self.layer = layer


class Field:
    def __init__(self, point, vectors, name=None, label=None, layer=None):
        self.point = point
        self.vectors = vectors
        self.name = name
        self.label = label
        self.layer = layer


class Mesh:
    def __init__(self, points, faces, name=None, label=None, layer=None):
        self.points = points
        self.faces = faces
        self.name = name
        self.label = label
        self.layer = layer



"""
A Point describe a position in a 3D space: location and orientation.
A Point can hold an assembly to specify the position of that assembly.

Manages reference frames and transformations.


Potential decision: 
- Pose(pose) copies the matrix


"""

import numpy as np
from scipy.spatial.transform import Rotation, Slerp

class Element:
    def __init__(self, name=None, label=None, layer=None):
        self.name = name
        self.label = label
        self.layer = layer

    def __getitem__(self, key):
        if key in self.__dict__:
            return self.__dict__[key]
        else:
            try:
               return getattr(self, key)
            except AttributeError:
                raise KeyError(f"{self} has no attribute {key}")

    def at(self, name=None, pose=None):
        if pose is None:
            return Pose(name=name, element=self)
        else:
            return pose.new(name=name,element=self)

    def render(self, style):
        return [Pose(name=self.name, element=self)]

    def draw2d(self, style=None, projection='xy'):
        from .canvas import Canvas2D
        canvas = Canvas2D(projection=projection, style=style)
        canvas.add(self)
        canvas.draw()
        return canvas


class Pose:
    def __init__(
        self,
        x=0,
        y=0,
        z=0,
        dx=[1, 0, 0],
        dy=[0, 1, 0],
        dz=[0, 0, 1],
        matrix=None,
        loc=None,
        rot=None,
        element=None,
        name=None,
        label=None,
        layer=None,
    ):
        if matrix is None:
            if loc is not None:
                x, y, z = loc
            if rot is not None:
                dx, dy, dz = rot.T
            self.matrix = np.array(
                [
                    [dx[0], dy[0], dz[0], x],
                    [dx[1], dy[1], dz[1], y],
                    [dx[2], dy[2], dz[2], z],
                    [0, 0, 0, 1],
                ],
                dtype=float,
            )
        elif isinstance(matrix, np.ndarray):
            self.matrix = matrix
        elif hasattr(matrix, "matrix"):
            self.matrix = matrix.matrix
        self.name = name
        self.label = label
        self.layer = layer
        self.element = element

    def __repr__(self):
        args = []
        if self.name is not None:
            args.append(f"{self.name!r}:")
        if self.element is not None:
            if self.element.name is not None:
                args.append(f"{self.element.name!r}")
            else:
                args.append(f"{self.element!r}")
        args.append(f"at {self.loc}")
        return f"<{' '.join(args)}>"

    def at(self, name, pose):
        return self.clone(name=name, matrix=pose.matrix@self.matrix)

    def __getitem__(self, path):
        path =path.split("/")
        key=path[0]
        name=f"{self.name}/{key}"
        res=self.element[key].at(name, pose=self)
        if len(path)>1:
            return res[path[1]]
        else:
            return res

    def __getattr__(self, key):
        try:
            return self[key]
        except AttributeError:
            raise AttributeError(f"{self} has no attribute {key}")

    @property
    def x(self):
        return self.matrix[0, 3]

    @x.setter
    def x(self, value):
        self.matrix[0, 3] = value

    @property
    def y(self):
        return self.matrix[1, 3]

    @y.setter
    def y(self, value):
        self.matrix[1, 3] = value

    @property
    def z(self):
        return self.matrix[2, 3]

    @z.setter
    def z(self, value):
        self.matrix[2, 3] = value

    @property
    def dx(self):
        return self.matrix[:3, 0]

    @dx.setter
    def dx(self, value):
        self.matrix[:3, 0] = value

    @property
    def dy(self):
        return self.matrix[:3, 1]

    @dy.setter
    def dy(self, value):
        self.matrix[:3, 1] = value

    @property
    def dz(self):
        return self.matrix[:3, 2]

    @dz.setter
    def dz(self, value):
        self.matrix[:3, 2] = value

    @property
    def loc(self):
        return self.matrix[:3, 3]

    @loc.setter
    def loc(self, value):
        self.matrix[:3, 3] = value

    @property
    def loc4(self):
        return self.matrix[:, 3]

    @property
    def rot(self):
        return self.matrix[:3, :3]

    @rot.setter
    def rot(self, value):
        self.matrix[:3, :3] = value

    @property
    def n(self):
        return self.new()

    @property
    def c(self):
        return self.clone()

    def __mul__(self, other):
        matrix = self.matrix.copy()
        matrix[:3, :3] = matrix[:3, :3] @ other.matrix[:3, :3]
        return self.clone(matrix=matrix)

    def __matmul__(self, other):
        matrix = self.matrix.copy()
        matrix = matrix @ other.matrix
        return self.clone(matrix=matrix)

    def __add__(self, other):
        matrix = self.matrix.copy()
        matrix[:3, 3] += other.matrix[:3, 3]
        return self.clone(matrix=matrix)

    def tx(self, x):
        self.matrix = np.dot(
            self.matrix,
            np.array([[1, 0, 0, x], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]),
        )
        return self

    def ty(self, y):
        self.matrix = np.dot(
            self.matrix,
            np.array([[1, 0, 0, 0], [0, 1, 0, y], [0, 0, 1, 0], [0, 0, 0, 1]]),
        )
        return self

    def tz(self, z):
        self.matrix = np.dot(
            self.matrix,
            np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, z], [0, 0, 0, 1]]),
        )
        return self

    def rx(self, angle):
        angle_rad = np.radians(angle)
        cx = np.cos(angle_rad)
        sx = np.sin(angle_rad)
        self.matrix = np.dot(
            self.matrix,
            np.array(
                [
                    [1, 0, 0, 0],
                    [0, cx, -sx, 0],
                    [0, sx, cx, 0],
                    [0, 0, 0, 1],
                ]
            ),
        )
        return self

    def ry(self, angle):
        angle_rad = np.radians(angle)
        cx = np.cos(angle_rad)
        sx = np.sin(angle_rad)
        self.matrix = np.dot(
            self.matrix,
            np.array(
                [
                    [cx, 0, sx, 0],
                    [0, 1, 0, 0],
                    [-sx, 0, cx, 0],
                    [0, 0, 0, 1],
                ]
            ),
        )
        return self

    def rz(self, angle):
        angle_rad = np.radians(angle)
        cx = np.cos(angle_rad)
        sx = np.sin(angle_rad)
        self.matrix = np.dot(
            self.matrix,
            np.array(
                [
                    [cx, -sx, 0, 0],
                    [sx, cx, 0, 0],
                    [0, 0, 1, 0],
                    [0, 0, 0, 1],
                ]
            ),
        )
        return self

    def clone(self, **kwargs):
        """Return a full clone of the current pose"""
        newargs = {**self.__dict__, **kwargs}
        return Pose(**newargs)

    def new(self, **kwargs):
        """Return a new pose with the same matrix as self"""
        return Pose(matrix=self.matrix, **kwargs)

    def distance(self, pose):
        return np.linalg.norm(self.matrix[:3, 3] - pose.matrix[:3, 3])

    def render(self, style):
        from .primitives import Text
        primitives = []
        if style.get("labels", False):
            primitives.append(
                Text(text=self.name).at(self,pose=self))
        if style.get("center.visible", True):
            primitives.append(self.new(name=self.name))
        if self.element is not None:
            for primitive in self.element.render(style):
                primitives.append(
                    primitive.at(name=f"{self.name}/{primitive.name}",pose=self)
                )
        return primitives

    def draw2d(self, style=None, projection='xy'):
        from .canvas import Canvas2D
        canvas = Canvas2D(projection=projection, style=style)
        canvas.add(self)
        canvas.draw()
        return canvas



class Frame(Element):
    def __init__(self, name, *parts, data=None, parent=None, prototype=None):
        self.name = name
        self.parts = {el.name: el for el in parts} # Poses of the parts
        self.data = data  # other metadata
        self.parent = parent  # name of the parent assembly
        self.prototype = prototype  # if it has been cloned

    def clone(self, **kwargs):
        data = {**self.__dict__, **kwargs}
        return super().__class__(prototype=self, **data)

    def __getitem__(self, path):
        path =path.split("/")
        key=path[0]
        res=self.parts[key]
        if len(path)>1:
            return res[path[1]]
        else:
            return res

    def render(self,style=None):
        primitives = []
        for part in self.parts.values():
            primitives.extend(part.render(style=style))
        return primitives




"""
name is for access
label is for presentation


element.render() -> generate list poses of primitives
canvas.draw_primitive() -> extract points or mesh from primitive and draw using style

"""

import matplotlib.pyplot as plt
import matplotlib as mpl
import matplotlib.lines as mlines
import matplotlib.patches as mpatches
import matplotlib.path as mpath


from matplotlib import patches
import numpy as np


def resolve_style(style, primitive, layer, name):
    if style is None:
        style = {}
    if layer in style:
        style = {**style, **style[layer]}
    if name in style:
        style = {**style, **style[name]}
    if primitive.__class__.__name__ in style:
        style = {**style, **style[primitive.__class__.__name__]}
    return style


class SimpleProjection:
    def __init__(self, axes="xy", scale=1, origin=[0, 0]):
        self.axes = axes
        if np.isscalar(scale):
            self.scale = [scale, scale]
        else:
            self.scale = scale
        self.origin = origin
        mapping = {"x": 0, "y": 1, "z": 2}
        self.idx0 = mapping[axes[0]]
        self.idx1 = mapping[axes[1]]

    def transform(self, points):
        """Transform points from 3D to 2D

        Args:
            points np.ndarray 3xN: N 3D points in columns
        """
        x = points[self.idx0] * self.scale[0] + self.origin[0]
        y = points[self.idx1] * self.scale[1] + self.origin[1]
        return x, y


class Canvas2D:
    default_style = {}

    def __init__(
        self,
        projection="xy",
        origin=[0, 0],
        scale=1,
        units="m",
        style=None,
        xlabel=None,
        ylabel=None,
        fig=None,
        ax=None,
    ):
        if isinstance(projection, str):
            self.projection = SimpleProjection(axes=projection, origin=origin, scale=scale)
        else:
            self.projection = projection
        if style is None:
            style = Canvas2D.default_style
        self.style = style
        self.units = units
        if xlabel is None:
            xlabel = f"{self.projection.axes[0]} [{self.units}]"
        self.xlabel = xlabel
        if ylabel is None:
            ylabel = f"{self.projection.axes[1]} [{self.units}]"
        self.ylabel = ylabel
        self.artists = {}
        self.elements = {}
        self.set_figure(fig, ax)

    def add(self, element):
        self.elements[element.name] = element
        return self

    def set_figure(self, fig, ax):
        if fig is not None:
            self.fig = fig
            if ax is None:
                self.ax = fig.gca()
            else:
                self.ax = ax
        if fig is None:
            if ax is None:
                self.fig, self.ax = plt.subplots()
                self.ax.set_aspect("equal")
                self.draw()
                self.fig.show()
            else:
                self.ax = ax
                self.fig = ax.get_figure()

    def clear(self):
        for key, artists in self.artists.items():
            for artist in artists:
                artist.remove()

    def draw(self, style=None):
        self.clear()
        self.ax.set_xlabel(self.xlabel)
        self.ax.set_ylabel(self.ylabel)
        if style is None:
            style = self.style
        for key, element in self.elements.items():
            for primitive in element.render(style):
                artists = self.draw_primitive(primitive, style)
                self.artists[key] = artists
        self.fig.show()

    def draw_primitive(self, primitive, style):
        style = resolve_style(
            style, primitive, primitive.layer, primitive.name
        )
        if style.get("visible", True):
            if primitive.element is None:
                artists=self.draw_pose(primitive, style)
            else:
                method=f"draw_{primitive.element.__class__.__name__}".lower()
                artists = getattr(self, method)( primitive, style)
            return artists
        else:
            return []

    def draw_pose(self, pose, style):
        x, y = self.projection.transform(pose.matrix[:3, 3])
        return [self.ax.text(x, y, pose.name)]

    def draw_line(self, primitive, style):
        points = primitive.matrix@primitive.element.points
        x, y = self.projection.transform(points)
        return [self.ax.plot(x, y, **style)]

    def draw_polyline(self, primitive, style):
        points = primitive.matrix@primitive.element.points
        x, y = self.projection.transform(points)
        return [self.ax.plot(x, y, **style)]

    def draw_polygon(self, primitive, style):
        points = primitive.matrix@primitive.element.points
        xy = np.array(self.projection.transform(points)).T
        patch = mpatches.Polygon(xy, edgecolor='k',facecolor='none', **style)
        print(xy.shape)
        self.ax.add_patch(patch)
        return [patch]



"""
The library models the layout of a beamline.

A Layout is a collections of elements in positions and orientations.

A Pose is a position and orientation in 3D space and can contain a reference to an element.

An element specify an object in space, could be primitive, such as a square or complex such as an assembly of elements.

An assembly is a container of parts specified by elements at a given pose.

pose[key] returns  Pose(element[key],name=f"{pose.name}/{key}").



An element has a render method that returns a list of primitives and their poses for diplsaying the element.


Points:
Single point (3,) ,(4.) or (3,1), (4,1)
Multipole pints (3,N), (4,N)

.points() -> return (4,N)
.poses() -> return list of poses
.point(s) -> for curves return Pose at s


Style:
  style is used by render to determine what primitives to emit draw
  style is used by draw_primitive to determine how to draw primitive 

"""

__version__="0.0.1"


from .assembly import Assembly, Magnet
from .layout import Beamline, Layout, Node, Env
from .pose import Pose, Element, Frame
from .primitives import (Bend, Box, Circle, Curve, Ellipse, Line, Polygon,
                         Polyline, Rectangle, Text, Tube)
from .canvas import Canvas2D



""" 
The assembly is a container of parts. A part is elemement/component at a given pose. 
"""

from typing import Any


class Assembly:
    @classmethod
    def from_yamldata(cls, name, data):
        kwargs = {}
        for attr in data:
            for kattr, vattr in attr.items():
                kwargs[kattr] = vattr

        return cls(name=name, **kwargs)

    def __init__(self, name=None, parts=None, data=None, parent=None, prototype=None):
        self.name = name
        self.parts = parts  # Poses with assemblies
        self.data = data  # other metadata
        self.parent = parent  # name of the parent assembly
        self.prototype = prototype  # if it has been cloned

    def at(self, point, name=None):
        return point.clone(type=self, name=name)

    def clone(self, **kwargs):
        data = {**self.__dict__, **kwargs}
        return super().__class__(prototype=self, **data)

    def __repr__(self):
        return f"{self.name}: {self.show_yaml()}"

    def show_yaml(self):
        attr = [str(self.prototype)]
        for k, v in self.__dict__.items():
            if k not in ["name", "parts", "prototype"]:
                if v is not None:
                    attr.append(f"{k}: {v}")
        return f"[{', '.join(attr)}]"


class Magnet(Assembly):
    def __init__(self, length=0, angle=0, tilt=0, aperture=None, name=None, parts=None):
        self.name = name
        self.length = length
        self.angle = angle
        self.tilt = tilt
        self.aperture = aperture
        super().__init__(name=name, prototype=self.__class__.__name__, parts=parts)


class Bend(Magnet):
    def __init__(self, length=0, angle=0, tilt=0, aperture=None, name=None):
        super().__init__(length, angle, tilt, aperture, name, parts=dict())


class Quadrupole(Magnet):
    def __init__(self, length=0, angle=0, tilt=0, aperture=None, name=None):
        super().__init__(length, angle, tilt, aperture, name, parts=dict())


class Region(Assembly):
    def __init__(self, length=0, angle=0, tilt=0, profiles=None, name=None, parts=None):
        self.name = name
        self.length = length
        self.angle = angle
        self.tilt = tilt
        self.profiles = profiles
        super().__init__(name=name, **parts)



"""

Nodes can be repeat in YAML but internally they are unique.


ll=xl.Layout.from_yaml('lattice1.yaml')
ll['MBB'][MB.1'] -> return the node



ll/'MBB/MB.1/center' -> returns a point object in the frame of the layout
ll/'MBB'/'MB.1'/'center' -> same as above?
ll['MBB']/'MB.1/center' -> returns a point object in the frame of MBB
ll['MBB'][MB.1'] -> return the node object
ll/'MBB'/'MB.1' -> return a point with the associated assembly

Questions:
ll['MBB/MB.1'] -> return the assembly object
ll['MBB/MB.1'] -> return point

"""

import yaml

from .pose import Pose
from .assembly import Assembly, Magnet, Bend, Quadrupole


class Node:
    def __init__(
        self,
        name,
        assembly,
        at=None,
        from_=None,
        ref="middle",
        ref_angle=None,
        ref_length=None,
        ref_roll=None,
        transform=None,
    ):
        self.name = name
        self.assembly = assembly
        self.at = at
        self.from_ = from_
        self.ref = ref
        self.ref_angle = ref_angle
        self.ref_length = ref_length
        self.ref_roll = ref_roll
        self.transform = transform
        if self.transform is None:
            self.transform = []
        if self.ref_angle is None:
            if hasattr(assembly, "angle"):
                self.ref_angle = assembly.angle
            else:
                self.ref_angle = 0
        if self.ref_length is None:
            if hasattr(assembly, "length"):
                self.ref_length = assembly.length
            else:
                self.ref_length = 0
        if self.ref_roll is None:
            if hasattr(assembly, "roll"):
                self.ref_roll = assembly.roll
            else:
                self.ref_roll = 0

    def __repr__(self):
        return f"{self.name}: {self.assembly}, at={self.at}"


class Segment:
    def __init__(self, length, angle, roll, start):
        self.length = length
        self.angle = angle
        self.roll = roll
        self.start = start

    def __repr__(self):
        return f"Segment: {self.length}, {self.angle}, {self.roll}, {self.start}"


class Beamline:
    @classmethod
    def from_yamldata(cls, name, data):
        nodes = {}
        for nodedata in data:
            [(nodename, attrs)] = nodedata.items()
            kwargs = {}
            transform = []
            assembly = attrs[0]
            for attr in attrs[1:]:
                for kattr, vattr in attr.items():
                    if kattr in ["tx", "ty", "tz", "rx", "ry", "rz"]:
                        transform.append([kattr, vattr])
                    else:
                        kwargs[kattr] = vattr
            kwargs["transform"] = transform
            kwargs["assembly"] = assembly
            kwargs["name"] = nodename
            nodes[nodename] = Node(**kwargs)
        return cls(name=name, nodes=nodes)

    def __init__(self, name=None, nodes=None):
        self.name = name
        self.nodes = nodes

    def find_sorted_nodes(self):
        abs_start = {}
        klist = set(self.nodes.keys())
        while len(klist) > 0:
            k = klist.pop()
            node = self.nodes[k]
            if node.at is not None:
                if node.from_ is None:
                    abs_start[k] = node.at
                elif k in abs_start:
                    abs_start[k] = abs_start[node.from_] + node.at
                else:
                    klist.add(k)
        sorted_nodes = sorted(abs_start.items(), key=lambda x: x[1])
        return sorted_nodes

    def find_segments(self):
        sorted_nodes = self.find_sorted_nodes()
        k, node_start = sorted_nodes.pop(0)
        node = self.nodes[k]
        cur_s = node_start + node.ref_length
        cur_angle = self.nodes[k].ref_angle
        cur_roll = self.nodes[k].ref_roll
        segments = [Segment(node.ref_length, cur_angle, cur_roll, node_start)]
        for k, node_start in self.find_sorted_nodes():
            node = self.nodes[k]
            if node_start < cur_s:  # node overlaps with previous node
                if node.ref_angle != cur_angle or node.ref_roll != cur_roll:
                    raise ValueError(
                        f"Node {k} overlaps with previous node but has different angle or roll"
                    )
                else:
                    node_end = node_start + node.ref_length
                    if node_end > cur_s:
                        segments[-1].length = node_end - cur_s
            cur_angle = node.ref_angle
            cur_roll = node.ref_roll
            cur_s += node.ref_length
            segments.append(Segment(node.ref_length, cur_angle, cur_roll, cur_s))
        return segments

    def __repr__(self):
        return f"{self.name}: {self.show_yaml()}"

    def show_yaml(self):
        if len(self.nodes) < 4:
            nodes = ", ".join(self.nodes)
        else:
            kk = list(self.nodes.keys())
            nodes = ", ".join(kk[:2] + ["..."] + kk[-2:])
        return f"[Sequence, {nodes}]"

    def __getitem__(self, key):
        return self.nodes[key]


class Layout:
    @classmethod
    def from_yaml(cls, filename):
        yamldata = yaml.safe_load(open("lattice1.yaml"))
        vars = {}
        data = {"vars": vars}
        for k, v in yamldata.items():
            if type(v) == str:  # variable definition
                vars[k] = v
            elif type(v) == list:  # assembly definition
                obj = assemblies[v[0]].from_yamldata(k, v[1:])
                data[k] = obj
        layout = cls(data)
        return layout

    def __init__(self, data):
        self.data = data

    def __repr__(self):
        return f"Layout: {len(self.data)-1} elements"

    def __getitem__(self, key):
        return self.data[key]

    def show(self):
        maxkey = max([len(k) for k in self.data.keys()])
        fmt = f"{{:<{maxkey}}}: {{}}"
        for k, v in self.data.items():
            if k != "vars":
                print(fmt.format(k, v.show_yaml()))


assemblies = {
    "Magnet": Magnet,
    "Bend": Bend,
    "Quadrupole": Quadrupole,
    "Assembly": Assembly,
    "Beamline": Beamline,
}



class Env:
    classes={}

    def __init__(self):
        self.elements = {}
        #self.mgr=xdeps.Manager()

    def add(self, element):
        if isinstance(element, str):
            element=self.parse(element)
        else:
            element=element

    def to_string(self):
        yield NotImplementedError
